#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <ctype.h>

/* buffer sizes */
#define BUFSIZE 256
#define BIGBUFSIZE 1024

/* filter options */
#define OP_NOCASE 1
#define OP_ABSOLUTE 2

/* mail-header information structure */
struct _MAILSTRUCT {
	char efrom[BUFSIZE];
	char wfrom[BUFSIZE];
	char subj[BUFSIZE];
};
typedef struct _MAILSTRUCT MAILSTRUCT;

/* filter chain structure */
typedef struct _FILTER FILTER;
typedef FILTER *PFILTER;
struct _FILTER {
	char f_fname[BUFSIZE];
	char f_text[BUFSIZE];
	char f_type[BUFSIZE];
	int f_options;
	PFILTER f_next;
};
	
/* folder chain structure */
typedef struct _FOLDER FOLDER;
typedef FOLDER *PFOLDER;
struct _FOLDER {
	char f_name[BUFSIZE];
	long f_offset;
	FILE *fp;
	PFOLDER f_next;
};

struct _PARSESTRUCT {
	char p_spoolpath[BUFSIZE];
	char p_stmppath[BUFSIZE];
	char p_confpath[BUFSIZE];
	char p_foldpath[BUFSIZE];
	char p_nofilpath[BUFSIZE];
	FILE *p_spool;
	FILE *p_nofilter;
	PFOLDER p_folders;
	PFILTER p_filters;
};
typedef struct _PARSESTRUCT PARSESTRUCT;

/* prototypes */
int copyspool(char *, char *); 
MAILSTRUCT *getmstruct(FILE *);
int copymessage(FILE *, FILE *);
int readconf(PARSESTRUCT *);
int addfolder(PFOLDER *, char *);
int addfilter(PFILTER *, char *, char *, char *, int);
int fexist(char *, PFOLDER *);
int filter(PARSESTRUCT *);
int cleanup(PARSESTRUCT *);
char *wclean(char *);
int getoptions(char *);
int respool(PARSESTRUCT *);
int init(PARSESTRUCT *);
int returnchar(int);
int mystrstr(char *, char*);
int checkfilter(PFILTER , MAILSTRUCT *, int (*)() ,int (*)() );
PFILTER getfilter(PFILTER , MAILSTRUCT *);
PFOLDER getfolder(PFOLDER , PFILTER );
int pabort(PARSESTRUCT *ps);


main(int argc, char *argv[])
{
	FILE *sfp, *mfp, *hfp;
	MAILSTRUCT *ms;
	PARSESTRUCT ps;
	long filelen;
	char test[256], *p;

	if(argc < 2){
		printf("Usage: pfilter <filterfile>\n");
		return(-1);
	}
	strncpy(ps.p_confpath,argv[1],BUFSIZE-1);
	printf("using %s\n",ps.p_confpath);
	if(init(&ps) == -1){
		printf("Error initializing program\n");
		abort(&ps);
		return(-1);
	}

	if(readconf(&ps) == -1) {
		abort(&ps);
		return(-1);
	}

/*	walkps(&ps); */  
	filter(&ps);
	cleanup(&ps);

}

/***************************************************/
/* FILTER - filters mail spool against filter list */
/***************************************************/
int filter(PARSESTRUCT *ps)
{
	
	char buf[BUFSIZE];
	MAILSTRUCT *ms;
	PFOLDER pfold;
	PFILTER pfil;

	while( (ms = getmstruct(ps->p_spool)) != NULL) {
		pfil = getfilter(ps->p_filters,ms);
		if(pfil != NULL)
			pfold = getfolder(ps->p_folders, pfil);
		if(pfil == NULL || pfold == NULL)
			copymessage(ps->p_spool, ps->p_nofilter);
		else 
			copymessage(ps->p_spool, pfold->fp);
	}
	return(0);

}

/***********************************************************/
/* GETFOLDER - locates filter struct folder in folder list */
/***********************************************************/
PFOLDER getfolder(PFOLDER top, PFILTER fil)
{
	PFOLDER cur;
	
	cur = top;
	while(cur != NULL){
		if(strcmp(cur->f_name,fil->f_fname) == 0)
			return(cur);
		else
			cur = cur->f_next;
	}
	return(NULL);
}
	
/*******************************************************/
/* GETFILTER - checks mail header against filter chain */
/*******************************************************/
PFILTER getfilter(PFILTER top, MAILSTRUCT *ms)
{
	PFILTER cur;
	int test;
	
	cur = top;
	while(cur != NULL) {
		switch(cur->f_options){
			case OP_NOCASE:
				test = checkfilter(cur, ms, toupper, mystrstr);
				break;
			case (OP_NOCASE | OP_ABSOLUTE):
				test = checkfilter(cur, ms, toupper, strcmp);
				break;
			case OP_ABSOLUTE:
				test = checkfilter(cur, ms, returnchar, strcmp);
				break;
			default:
				test = checkfilter(cur, ms, returnchar, mystrstr);
		}
		if(test == 0)
			return(cur);
		else
			cur = cur->f_next;
	}
	return(NULL);
}/* end getfilter */


/*****************************************************/
/* CHECKFILTER - checks if mailheader matches filter */
/*****************************************************/
int checkfilter(PFILTER fil, MAILSTRUCT *ms, int (*charop)(),int (*compare)() )
{
	char fbuf[BUFSIZE], mbuf[BUFSIZE];
	int i;

	strncpy(fbuf,fil->f_text,BUFSIZE-1);
	if(strcmp(fil->f_type,"EMAIL") == 0)
		strncpy(mbuf,ms->efrom,BUFSIZE-1);
	else if(strcmp(fil->f_type,"NAME") == 0)
		strncpy(mbuf,ms->wfrom,BUFSIZE-1);
	else if(strcmp(fil->f_type,"SUBJECT") ==0)
		strncpy(mbuf,ms->subj,BUFSIZE-1);
	else
		return(-1);
	
	for(i = 0; i < strlen(fbuf); i++)
		fbuf[i] = (*charop)(fbuf[i]);
	for(i = 0; i < strlen(mbuf); i++)
		mbuf[i] = (*charop)(mbuf[i]);
	
	i = (*compare)(fbuf,mbuf);
	return(i);
}
/* end checkfilter */	

/*****************************************************/
/* MYSTRSTR - returns success or failure of strstr() */		
/*****************************************************/
int mystrstr(char *s, char *t)
{
	if(strstr(t,s) == NULL)
		return(-1);
	else
		return(0);
}

/********************************************************/
/* RETURNCHAR - dummy function to call from checkfilter */
/********************************************************/
int returnchar(int c)
{
	return(c);
}

/**********************************/
/* INIT - initializes parsestruct */
/**********************************/
int init(PARSESTRUCT *ps)
{
	umask(0);

	ps->p_spool = NULL;
	ps->p_nofilter = NULL;
	ps->p_folders = NULL;
	ps->p_filters = NULL;
	strcpy(ps->p_nofilpath,"");
	strcpy(ps->p_foldpath,"");
	strcpy(ps->p_spoolpath,"");
	strcpy(ps->p_stmppath,"");

	/* open catchall file */
	strncpy(ps->p_nofilpath,"/tmp/pf1XXXXXX",BUFSIZE-1); 
	mktemp(ps->p_nofilpath);             /* make catch-all folder */
	ps->p_nofilter = fopen(ps->p_nofilpath,"w+");
	if(ps->p_nofilter == NULL){
		printf("Error: cannot open temp file. aborting\n");
		cleanup(ps);
		return(-1);
	}
}/* end init */

/*************************************/
/* ABORT - replace spool and exit    */
/*************************************/
int pabort(PARSESTRUCT *ps)
{

	fclose(ps->p_spool);
	if(copyspool(ps->p_stmppath, ps->p_spoolpath) == -1)
		return(-1);
	cleanup(ps);
}	

/*********************************************/
/* CLEANUP - closes files, frees parsestruct */
/*********************************************/
int cleanup(PARSESTRUCT *ps)
{
	PFOLDER cfold,nfold;
	PFILTER cfil,nfil;

	cfold = ps->p_folders;   /* close all folders */
	while(cfold != NULL){
		fclose(cfold->fp);
		cfold = cfold->f_next;
	}

	cfold = ps->p_folders;   /* free folder structs */
	while(cfold != NULL){
		nfold = cfold->f_next;
		free(cfold);
		cfold = nfold;
	}

	cfil = ps->p_filters;   /* free filter structs */
	while(cfil != NULL){
		nfil = cfil->f_next;
		free(cfil);
		cfil = nfil;
	}
	fclose(ps->p_spool);
	remove(ps->p_stmppath);

	if(ps->p_nofilter != NULL) {
		respool(ps);
		fclose(ps->p_nofilter);
		remove(ps->p_nofilpath);
	}
}/* end cleanup */

/***************************************/
/* RESPOOL - respools unfiltered spool */
/***************************************/
int respool(PARSESTRUCT *ps)
{
	int spoolfd;
	char buf[BUFSIZE];

	rewind( ps->p_nofilter);
	spoolfd = open( ps->p_spoolpath,O_WRONLY | O_APPEND);
	if(spoolfd < 0)
		perror(ps->p_spoolpath);
	while(flock(spoolfd,LOCK_EX) == -1)
		sleep(5);
	while( fread(buf,BUFSIZE-1,1,ps->p_nofilter) > 0)
		write(spoolfd,buf,BUFSIZE-1);
	flock(spoolfd,LOCK_UN);
	close(spoolfd);
	return(0);
}
/* end respool */

/**************************************/
/* WALKPS  prints out the parsestruct */
/**************************************/
int walkps(PARSESTRUCT *ps)
{
	PFOLDER fold;
	 PFILTER fil;
	
	printf("%s\n",ps->p_spoolpath);
	printf("%s\n",ps->p_stmppath);
	printf("%s\n",ps->p_confpath);
	
	fold = ps->p_folders;
	fil = ps->p_filters;
	
	printf("Open Folders...\n");
	while(fold != NULL){
		printf("\t%s : %d\n",fold->f_name,fold->f_offset);
		fold = fold->f_next;
	}
	printf("\tNULL\n\n");
	
	printf("Filters...\n");
	while(fil != NULL){
		printf("\t%s : %s : %s : %d\n",fil->f_type, fil->f_text, fil->f_fname, fil->f_options);
		fil = fil->f_next;
	}
	printf("\tNULL\n\n");
}

	
/**********************************************************/
/* READCONF - reads in config file and uses it to fill ps */
/**********************************************************/
int readconf(PARSESTRUCT *ps)
{
	FILE *config;
	char buf[BIGBUFSIZE], token[BIGBUFSIZE], *value;
	char folder[BUFSIZE],text[BUFSIZE],type[BUFSIZE];
	int options;

	config = fopen(ps->p_confpath,"r");
	if(config == NULL) {
		printf("Error: cannot read config file %s\n",ps->p_confpath);
		return(-1);
	}


	while( (fgets(buf,BIGBUFSIZE-1,config)) != NULL) {
		if(buf[0] == '#')          /* skip comments */
			continue;
		if(buf[0] == '\n')         /* skip empty lines */
			continue;
		strncpy(token,buf,BIGBUFSIZE-1);
		value = strtok(token,",");
		if(value == NULL)				/* skip blank lines */
			continue;	
		value = wclean(value);
		if(strcmp(value,"mailspool") == 0){         /* get mailspool path */
			if( (value = strtok(NULL,"\n")) == NULL)
				continue;
			value = wclean(value);
			strncpy(ps->p_spoolpath,value,BUFSIZE-1);
			ps->p_spoolpath[BUFSIZE-1] = '\0';
		}else if(strcmp(value,"mailfolder") == 0){  /* get mailfolder path */
			if( (value = strtok(NULL,"\n")) == NULL)
				continue;
			value = wclean(value);
			strncpy(ps->p_foldpath,value,BUFSIZE-1);
			ps->p_foldpath[BUFSIZE-1] = '\0';
		chdir(ps->p_foldpath);
		}else if(strcmp(value,"filter") == 0) {
			if( (value = strtok(NULL,",")) == NULL)  /*get filtertype */
				continue;
			else if(strchr(value,'\n')){
				printf("error reading filterfile at token %s\n",value);
				return(-1);
			}
			strncpy(type,wclean(value),BUFSIZE-1);
			type[BUFSIZE-1] = '\0';

			if( (value = strtok(NULL,",")) == NULL)  /* get filtertext */
				continue;
			else if(strchr(value,'\n')){
				printf("error reading filterfile at token %s\n",value);
				return(-1);
			}
			strncpy(text,wclean(value),BUFSIZE-1);
			text[BUFSIZE-1] = '\0';

			if( (value = strtok(NULL,",")) == NULL)  /*get folder */
				continue;
			else if(strchr(value,'\n')){
				printf("error reading filterfile at token %s\n",value);
				return(-1);
			}
			strncpy(folder,wclean(value),BUFSIZE-1);
			folder[BUFSIZE-1] = '\0';

			if( (value = strtok(NULL,"\n")) == NULL)  /*get options */
				options = 0;
			else
				options = getoptions(wclean(value));
			addfilter(&(ps->p_filters),folder,text,type,options); 
			if(fexist(folder,&(ps->p_folders)) == 0)       /* add to chains */       
				addfolder(&(ps->p_folders),folder);
		}
			
	}

	strncpy(ps->p_stmppath,"/tmp/pf2XXXXXX",BUFSIZE-1);
	mktemp(ps->p_stmppath);
	copyspool(ps->p_spoolpath,ps->p_stmppath);
	ps->p_spool = fopen(ps->p_stmppath,"r");
	if(ps->p_spool == NULL)
		return(-1);
	
	return(1);
} /* end readconf */


/*******************************************************/
/* GETOPTIONS - converts option string to option value */
/*******************************************************/
int getoptions(char *ops)
{
	int options = 0;

	while(ops[0] != (char)NULL){
		switch(ops[0]){
			case 'C':	
				options = options | OP_NOCASE;
				break;
			case 'a':
				options = options | OP_ABSOLUTE;
				break;
			default:
		}
		ops++;
	}
	return(options);
} /* end getoptions */

/**********************************************************/
/* WCLEAN - returns s without leading/trailing whitespace */
/**********************************************************/
char *wclean(char *s)
{
	int i = 0;
	while(s[0] == ' ')
		s++;
	i = strlen(s) - 1;
	if(s[i] == '\n')
		s[i] == '\0';
	while(s[--i] == ' ')
		s[i] = '\0';
	return(s);
} /* end wclean */
		
/********************************************************************/
/* FEXIST - checks if the folder already exists in the folder chain */
/********************************************************************/
int fexist(char *name, PFOLDER *top)
{
	PFOLDER cur;
	char fname[BIGBUFSIZE];

	cur = *top;
	while(cur != NULL) {
		if(strcmp(name, cur->f_name) == 0)
			return(1);
		else {
			getcwd(fname,BUFSIZE-1);
			strncat(fname,name,BUFSIZE);	
			if(strcmp(fname, cur->f_name) == 0)
				return(1);
		}
		cur = cur->f_next;
	}
	return(0);
} /* end fexist */


/*********************************************************/
/* ADDFOLDER - adds a FOLDER to the end of a FOLDER list */
/*********************************************************/
int addfolder(PFOLDER *top, char *name)
{
	PFOLDER newf, prev;
	int i;

	(PFILTER)newf = malloc(sizeof(FOLDER));
	if(newf == NULL){
		printf("Error: Out of memory\n");
		free(newf);
		return(-1);
	}
	strncpy(newf->f_name,name,BUFSIZE);
	newf->f_name[BUFSIZE-1] = '\0';
	if(access(newf->f_name,F_OK) != 0){
		i = creat(newf->f_name,S_IRUSR | S_IWUSR);
		close(i);
	}
	newf->fp = fopen(newf->f_name,"a");
	if(newf->fp == NULL){
		printf("Error: cannot open %s for writing\n",newf->f_name);
		free(newf);
		return(-1);
	}
	newf->f_offset = 0;

	if(*top == NULL)
		*top = newf;
	else {
		prev = *top;
		while(prev->f_next != NULL)
			prev = prev->f_next;
		prev->f_next = newf;
	}
	newf->f_next = NULL;

	return(1);
} /* end addfolder */


/*********************************************************/
/* ADDFILTER - adds a FILTER to the end of a FILTER list */
/*********************************************************/
int addfilter(PFILTER *top, char *fname, char *text, char *type, int option)
{
	PFILTER newf, prev;

	(PFILTER)newf = malloc(sizeof(FILTER));
	if(newf == NULL) {
		printf("Error: Out of memory\n");
		free(newf);
		return(-1);
	}
	strncpy(newf->f_fname,fname,BUFSIZE);
	strncpy(newf->f_text,text,BUFSIZE);
	strncpy(newf->f_type,type,BUFSIZE);
	newf->f_options = option;
	newf->f_fname[BUFSIZE-1] = '\0';
	newf->f_text[BUFSIZE-1] = '\0';
	newf->f_type[BUFSIZE-1] = '\0';

	if(*top == NULL)
		*top = newf;
	else {
		prev = *top;
		while(prev->f_next != NULL)
			prev = prev->f_next;
		prev->f_next = newf;
	}
	newf->f_next = NULL;
	return(1);
} /* end addfilter */


/******************************************************************/
/* COPYMESSAGE - copies next mail message from spool to dest file */
/******************************************************************/
int copymessage(FILE *spool, FILE *dest)
{
	char buf[BIGBUFSIZE];
	int moved = 0;
	
	if(fgets(buf,BIGBUFSIZE-1,spool) == NULL)
		return(-1);
	if(strncmp(buf,"From ",5) != 0) {
		fseek(spool, -strlen(buf),SEEK_CUR);
		return(-1);
	}
	fputs(buf,dest);
	fgets(buf,BIGBUFSIZE-1,spool);
	while(strncmp(buf,"From ",5) != 0) {
		fputs(buf,dest);
		if(fgets(buf,BIGBUFSIZE-1,spool) == NULL)
			return(1);
	}
	fseek(spool,-strlen(buf),SEEK_CUR);
	return(1);
}		
/* end copymessage */

	
/************************************************************/
/* GETMSTRUCT - reads MAILSTRUCT info from next mail header */
/************************************************************/
MAILSTRUCT *getmstruct(FILE *spool)
{
	MAILSTRUCT *ms;
	long offset;
	char buf[BUFSIZE] = {0};
	int i;

	offset = ftell(spool);
	ms = (MAILSTRUCT *)malloc(sizeof(MAILSTRUCT));
	
	while(buf[0] != '\n') {
		if(fgets(buf,BUFSIZE-1,spool) == NULL){     /* we should never hit EOF */
			free(ms);
			return(NULL);
		}
		if(strncmp(buf,"From ",5) == 0)           /* get from field */
			if(strncpy(ms->efrom,&buf[5],BUFSIZE)) {
				ms->efrom[BUFSIZE-1] = '\0';
				ms->efrom[strlen(ms->efrom) - 1] = '\0';
			}
		if(strncmp(buf,"From: ",6) == 0)         /* get name field */
			if( strncpy(ms->wfrom,&buf[6], BUFSIZE) ) {
				ms->wfrom[BUFSIZE-1] = '\0';
				ms->wfrom[strlen(ms->wfrom) - 1] = '\0';
			}
		if(strncmp(buf,"Subject: ",9) == 0)       /* subject field */
			if( strncpy(ms->subj,&buf[9], BUFSIZE) ) {
				ms->subj[BUFSIZE-1] = '\0';
				ms->subj[strlen(ms->subj) - 1] = '\0';
			}
	}

	i = 0;
	while(ms->efrom[i] != ' ')
		i++;
	ms->efrom[i] = '\0';

	fseek(spool, offset, SEEK_SET);
	return(ms);
}
/* end getmstruct */


/***************************************/
/* COPYSPOOL - make temp copy of spool */
/***************************************/
int copyspool(char *spath, char *slocal)
{
	int sfd, lfd;
	int ret = 0;
	char buf[BUFSIZE];

	if( (sfd = open(spath,O_RDONLY)) == -1)
		return(-1);
	while(flock(sfd,LOCK_EX) == -1)
		sleep(5);

	
	if( (lfd = open(slocal,O_WRONLY | O_CREAT | O_APPEND,S_IRUSR | S_IWUSR))
			 == -1){
		ret = -1;
		goto done;
	}
	while(flock(sfd,LOCK_EX) == -1)
		sleep(5);	
	
	while( read(sfd,buf,BUFSIZE-1) > 0)
		write(lfd, buf,strlen(buf));

done:
	flock(sfd,LOCK_UN);
	flock(lfd,LOCK_UN);
	close(sfd);
	close(lfd);

	truncate(spath,0);
	
	return(ret);
}
/* end copyspool */
